/*  

columnize version 1.1, 4/9/98
by aaron brick, aaron@jhu.edu

please see the other files in
the distribution for information.

*/

#include <stdio.h>
#include <unistd.h>

#define TRUE 1
#define FALSE 0

/* directions, modes, version */

#define VERTICAL 0
#define HORIZONTAL 1

#define LEFT 0
#define RIGHT 1

#define NORMAL 0
#define PEDANTIC 1
#define DIAGNOSTIC 2
#define LS_EMULATION 3

#define NOT_SPREAD 0
#define SPREAD 1

#define VERSION_A 1
#define VERSION_B 1

/* default values */

#define DEFAULT_SW 80
#define DEFAULT_CS 1
#define DEFAULT_DIR VERTICAL
#define DEFAULT_MODE NORMAL
#define DEFAULT_STYLE NOT_SPREAD
#define DEFAULT_JUS LEFT
#define MAX_LINE_LENGTH 1025

/* linked list for dynamic storage of lines */

typedef struct unit
{
	char *text;
	struct unit *next;
} data;

int dnru ( int,int );           /* "divide 'n' round up" */
void spew ( data* );            /* spew out input (fast) */
void error ( int,int,int );     /* gives error msg and quits */
void help ();                   /* prints out help and quits */
int printable_length ( char* ); /* the actual printable length of a string */

void main ( int argc, char **argv )
{
	int screenwidth = DEFAULT_SW, cellspacing = DEFAULT_CS, justification = DEFAULT_JUS;
	int direction = DEFAULT_DIR, mode = DEFAULT_MODE, style = DEFAULT_STYLE;
	int numelts = 0, longest = 0, numcols, fullcols, numlines, numspaces, lastcols, requestedcols = 0;
	int loop, drawlines, drawelts, drawspaces, index, read = 0, len, datlen, req = FALSE;
	int legit = 0;
	char this;
	long memory = 0;

	data *item, *first;
	char temp [ MAX_LINE_LENGTH ], temp2 [ MAX_LINE_LENGTH ], ciq, **line;

	if ( getenv ( "COLUMNS" ) )
		screenwidth = atoi ( getenv ( "COLUMNS" ) );

	/* parse arguments (or have getopt do it for me) */
	
	opterr = 0;

	while ( ( this = getopt ( argc, argv, ":vznpdlsufrw:c:x:h") ) != -1 )
		switch ( this )
		{
			case 'v': direction = VERTICAL ; break;
			case 'z': direction = HORIZONTAL ; break;
			
			case 'n': mode = NORMAL ; break;
			case 'p': mode = PEDANTIC ; break;
			case 'd': mode = DIAGNOSTIC ; break;
			case 'l': mode = LS_EMULATION ; break;
			
			case 's': style = SPREAD ; break;
			case 'u': style = NOT_SPREAD ; break;
			
			case 'w': screenwidth = atoi ( optarg ); break;
			case 'c': cellspacing = atoi ( optarg ); break;
			case 'x': requestedcols = atoi ( optarg ); req = TRUE; break;
			
			case 'f': justification = LEFT; break;
			case 'r': justification = RIGHT; break;
	
			case 'h': help (); break;
			case ':': error ( 1, mode, optopt ); break;
			case '?': error ( 2, mode, optopt ); break;
		}

	if ( screenwidth < 1 ) error ( 3, mode, screenwidth );
	if ( cellspacing < 0 ) error ( 4, mode, cellspacing );
	if ( ( req == TRUE ) && ( requestedcols < 2 ) ) error ( 0, mode, requestedcols );
	
	if ( mode == LS_EMULATION )
	{
		cellspacing = 2;
		direction = VERTICAL;
		justification = LEFT;
	}

	/* get input into linked list */
 
	item = (data*) malloc ( sizeof ( data ) );
	first = item;

	datlen = sizeof ( data );

	while ( fgets ( temp2, MAX_LINE_LENGTH + 1, stdin ) != NULL )
	{
		strncpy ( temp, temp2, strlen ( temp2 ) - 1 );
		temp [ strlen ( temp2 ) - 1 ] = '\0';
		
		len = strlen ( temp ) + 1;
		memory += ( len + datlen + 4 );
	
		item->text = (char*) malloc ( len );
		strcpy ( item->text, temp );
		
		item->next = (data*) malloc ( datlen );
		item = item->next;
		
		numelts++;
	}
	
	if ( numelts == 0 ) error ( 5, mode, 0 );

	/* preparing array of pointers */

	line = (char**) calloc ( numelts, sizeof (char*) );
	item = first;

	/* indexing linked list with array of pointers */

	for ( loop = 0 ; loop < numelts ; loop++ )
	{
		line[loop] = item->text;
		
		if ( printable_length ( item->text ) > longest )
			longest = printable_length ( item->text );
			
		item = item->next;
	}

	/* determine # of columns */

	fullcols = ( ( screenwidth + cellspacing ) / ( longest + cellspacing ) );
	
	if ( requestedcols > 1 )
	{
		if ( requestedcols > fullcols )
			error ( 8, mode, requestedcols );
		fullcols = requestedcols;
	}

	/* more calculations */

	if ( fullcols > 1 )
	{
		if ( ( numelts % fullcols ) == 0 )
			lastcols = fullcols;
		else
			lastcols = ( numelts % fullcols );

		numlines = dnru ( numelts, fullcols );
	}
	
	else if ( mode != DIAGNOSTIC )
	{
		if ( fullcols == 1 ) { error ( 7, mode, 0 ); spew ( first ); }
		if ( fullcols == 0 ) { error ( 6, mode, 0 ); spew ( first ); }
	}
	
	if ( style == SPREAD ) cellspacing += ( ( screenwidth - ( ( ( longest * fullcols ) + ( ( fullcols - 1 ) * cellspacing ) ) ) ) / ( fullcols - 1 ) );
	
	/* diagnostic section */
	
	if ( mode == DIAGNOSTIC )
	{
		printf ( "cols v%d.%d, by aaron@jhu.edu\n", VERSION_A, VERSION_B );
		printf ( "screenwidth is %d and cellspacing is %d.\n", screenwidth, cellspacing );
		printf ( "%d strings with a max length of %d were read.\n", numelts, longest );
		printf ( "%d bytes of memory were used storing the data.\n", memory );
		
		printf ( "expect ");
		if ( fullcols > 1 )
			if ( fullcols == lastcols ) printf("%d lines of %d elements each.\n", numlines, fullcols) ;
			else printf ( "%d lines of %d elements each and 1 line of %d element(s).\n", numlines-1, fullcols, lastcols );
		else
			printf ( "%d lines of 1 element each.\n", numelts );
		exit ( 0 );
	}

	/* print out columns */
	
	for ( drawlines = 0 ; drawlines < numlines ; drawlines++ )
	{
		/* setting whether this line is full-length or the last line */
	
		if ( drawlines == ( numlines - 1 ) )
			numcols = lastcols;
		else
			numcols = fullcols;

		for ( drawelts = 0 ; drawelts < numcols ; drawelts++ )
		{
			/* the guts of the algorithm */

			if ( direction == HORIZONTAL )
				index = ( drawelts + ( fullcols * drawlines ) );
				
			if ( direction == VERTICAL )
			{
				index = drawlines;
				if ( drawelts < lastcols )
					index += ( drawelts * numlines );
				else
					index += ( lastcols * numlines ) + ( ( drawelts - lastcols ) * ( numlines - 1 ) );
			}
			
			/* printing */
			
			/* replace tabs with spaces (they fuck shit up) */

			for ( loop = 0 ; loop < strlen ( line [ index ] ) ; loop++ )
				if ( (int) line [ index ] [ loop ] == 9 )
					line [ index ] [ loop ] = ' ';
			
			if ( justification == RIGHT )
				for ( drawspaces = printable_length ( line [ index ] ) ; drawspaces < longest ; drawspaces++ )
					printf ( " " );

			printf( "%s", line [ index ] );
			
			/* does it need spaces or a return? */
			
			if ( ( ( drawelts + 1 ) % numcols ) == 0 )
				printf ( "\n" );			
			else
			{
				if ( justification == LEFT )
					numspaces = ( longest - printable_length ( line [ index ] ) ) + cellspacing;
				else
					numspaces = cellspacing;
					
				for ( drawspaces = 0 ; drawspaces < numspaces ; drawspaces++ )
					printf ( " " );
			}
		}
	}
}

/* a division rounded up for numlines */

int dnru ( int a, int b )
{
	int num = 0, times = 1;

	while ( ( num += b ) < a )
		times++;
		
	return times;
}

/* for when it's faster to do this than use
   the formal algorithm */

void spew ( data *item )
{
	while ( item->next != NULL )
		printf  ( "%s\n", item->text );
		
	exit ( 0 );
}

/* for error reporting and/or quitting */

void error ( int code, int mode, int arg )
{
	if ( code <= 4 )
	{
		switch ( code )
		{
			case 1: fprintf ( stderr, "the option -%c requires a numeric argument\n", arg); break;
			case 2: fprintf ( stderr, "option -%c unrecognized; use -h for help\n", arg); break;
			case 3: fprintf ( stderr, "%d is a unacceptable screenwidth (must be > 0)\n", arg); break;
			case 4: fprintf ( stderr, "%d is a unacceptable cellspacing (must be >= 0)\n", arg); break;
			case 0: fprintf ( stderr, "%d is a unacceptable number of columns (must be > 1)\n", arg); break;
		}
		exit ( -1 );
	}
	else if ( mode != PEDANTIC )
		switch ( code )
		{
			case 5: exit (-1);
			case 6: return;
			case 7: return;
		}
	else
	{
		switch ( code )
		{
			case 5: fprintf ( stderr, "no input\n" ); break;
			case 6: fprintf ( stderr, "screenwidth is shorter than the longest line of input\n" ); break;
			case 7: fprintf ( stderr, "only one column would be possible\n" ); break;
			case 8: fprintf ( stderr, "%d columns won't fit\n", arg ); break;
		}
		exit ( -1) ;
	}
}

/* prints out the help */

void help ()
{
	printf ( "cols v%d.%d, by aaron@jhu.edu\n", VERSION_A, VERSION_B );
	printf ( "Usage: cols [-v|-z] [-n|-p|-d|-l] [-s|-u] [-f|-r] [-w n] [-c n] [-x n] [-h]\n\n" );

	printf ( "  -n    normal mode (output) [default]\n" );
	printf ( "  -p    pedantic mode (output, errors)\n" );
	printf ( "  -d    diagnostic mode (errors, information)\n" );
	printf ( "  -l    ls emulation mode\n\n" );
	
	printf ( "  -v    vertical output [default]\n" );
	printf ( "  -z    horizontal output\n\n" );

	printf ( "  -f    left justify columns [default]\n" );
	printf ( "  -r    right justify columns\n\n" );

	printf ( "  -s    spread columns to screen width\n" );
	printf ( "  -u    don't spread columns [default]\n\n" );

	printf ( "  -w n  set screenwidth to n characters [default n = 80]\n" );
	printf ( "        n must be > 0.\n" );
	printf ( "  -c n  set cellspacing to n characters [default n = 1]\n" );
	printf ( "        n must be >= 0.\n" );
	printf ( "  -x n  output in n columns [default n = maximum]\n" );
	printf ( "        n must be > 1.\n\n" );
	
	printf ( "  -h    display this help\n" );
	
	exit ( 0 );
}

/* returns actual printable length of string */

int printable_length ( char *string )
{
	int i = 0, j = 0;
	int a, ciq, nciq;
	
	while ( i < strlen ( string ) )
	{
		ciq = (int) string [i];
		nciq = (int) string [i + 1];
		
		/* start process if we see a control character */
			
		if ( ( ( ciq <= -97 ) || ( ( ciq <= 31 ) && ( ciq >= 0 ) ) || ( ( ciq <= 160 ) && ( ciq >= 128 ) ) ) && ( ciq != 9 ) )
		{
			if ( ( ciq == 27 ) && ( nciq == 91 ) )
			{
				a = i + 2;
				while ( ( ( string [a] >= '0' ) && ( string [a] <= '9' ) ) || ( string [a] == ';' ) )
					a++;

				if ( (int) string [a] == 109 )
					i = a;
			}
		}
			
		else
			j++;

		i++;
	}
	
	return j;
}
